% MC-LHS variance estimation (ROSA-VSI)
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%   H.-X. Zhang, W.P. Dempsey, Jr., and J. Goutsias    %%%%%%%%%%%
%%%%%%%%%%        Probabilistic Sensitivity Analysis of         %%%%%%%%%%%
%%%%%%%%%%            Biochemical Reaction Systems              %%%%%%%%%%%
%%%%%%%%%%  The Journal of Chemical Physics, 131, 094101, 2009  %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%        Reaction-Oriented Sensitivity Analysis        %%%%%%%%%%%
%%%%%%%%%%                   VSI COMPUTATION                    %%%%%%%%%%%
%%%%%%%%%%              Last Modified: 08/11/2009               %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Initializations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

perm1 = repmat(0,[M,L]);
perm2 = repmat(0,[M,L]);
YR1   = repmat(0,[2*M,1]);
YR2   = repmat(0,[2*M,1]);
YR3   = repmat(0,[M,2*M]);
YR4   = repmat(0,[M,2*M]);

meanTT(1:M) = 0;
meanDD(1:M) = 0;
meanSS(1:M) = 0;

varmTT(1:M) = 0;
varmDD(1:M) = 0;
varmSS(1:M) = 0;

varmmTT(1:M,1:M) = 0;
varmmDD(1:M,1:M) = 0;
varmmSS(1:M,1:M) = 0;

varfTT(1:M) = 0;
varfDD(1:M) = 0;
varfSS(1:M) = 0;

varvTT(1:M) = 0;
varvDD(1:M) = 0;
varvSS(1:M) = 0;

var2fTT(1:M,1:M) = 0;
var2fDD(1:M,1:M) = 0;
var2fSS(1:M,1:M) = 0;

rand('state',sum(100*clock));   % initialize random number generator

constr = log((h/(k*T)).*kappa); % standard chemical potential variation constraint
  
LL = L;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% MC-LHS iterations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Determine LHS permutations
%

for m = 1:M
    perm1(m,:) = randperm(L);
    perm2(m,:) = randperm(L);
end

for iter = 1:L
    
    iterleft = L+1-iter;
    fprintf('Remaining MC-LHS iterations: %6.0f\n',iterleft)

%
% Calculate Latin hypercube samples
%

    uniform = rand(M,1);
    index   = (perm1(:,iter)-1+uniform)./L;
    YR1     = (rlambda*(-sqrt(2)*erfcinv(2*index)))';  % inverse normal CDF
    temp    = repmat(YR1,[2 1]);      % replicate each element of YR1 twice
    YR1     = temp(:).';              % replicate each element of YR1 twice
    
    if YR1 >= constr'        % check standard chemical potential constraint
    else
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end

    uniform = rand(M,1);
    index   = (perm2(:,iter)-1+uniform)./L;
    YR2     = (rlambda*(-sqrt(2)*erfcinv(2*index)))';  % inverse normal CDF
    temp    = repmat(YR2,[2 1]);      % replicate each element of YR2 twice
    YR2     = temp(:).';              % replicate each element of YR2 twice

    if YR2 >= constr'        % check standard chemical potential constraint
    else
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end
    
    for m = 1:M
        YR3(m,1:2*M) = YR2(1:2*M);
        YR4(m,1:2*M) = YR1(1:2*M);
    end

    for m = 1:M
        YR3(m,2*m-1) = YR1(2*m-1);
        YR3(m,2*m)   = YR1(2*m);
        YR4(m,2*m-1) = YR2(2*m-1);
        YR4(m,2*m)   = YR2(2*m);
    end
    
%
% Calculate response characteristics (timing, duration, and strength)
%
% For kappa1
 
    kappa1 = kappa.*exp(-YR1)';
    if iode == 1
        [t,q] = mapk(kappa1,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa1,stoich);
    end
    [TT1,DD1,SS1] = featureit(species,q,t,tresol,durtol);
   
    %
    % For kappa2
    %
    
    kappa2 = kappa.*exp(-YR2)';
    if iode == 1
        [t,q] = mapk(kappa2,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa2,stoich);
    end
    [TT2,DD2,SS2] = featureit(species,q,t,tresol,durtol);
     
    %
    % For kappa3 & kappa4
    %
    
    for m = 1:M
        
        kappa3 = kappa.*exp(-YR3(m,:))';
        if iode == 1
            [t,q] = mapk(kappa3,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa3,stoich);
        end
        [TT3(m),DD3(m),SS3(m)] = featureit(species,q,t,tresol,durtol);
    
        kappa4 = kappa.*exp(-YR4(m,:))';
        if iode == 1
            [t,q] = mapk(kappa4,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa4,stoich);
        end
        [TT4(m),DD4(m),SS4(m)] = featureit(species,q,t,tresol,durtol);
   
    end

    %
    % Update mean and variance calculations
    %
    
    meanTT = meanTT + TT1*TT2 + TT3.*TT4;
    meanDD = meanDD + DD1*DD2 + DD3.*DD4;
    meanSS = meanSS + SS1*SS2 + SS3.*SS4;
    
    varmTT = varmTT + (TT1-TT2)^2 + (TT3-TT4).^2;
    varmDD = varmDD + (DD1-DD2)^2 + (DD3-DD4).^2;
    varmSS = varmSS + (SS1-SS2)^2 + (SS3-SS4).^2;
    
    for m = 1:M
        varmmTT(m,1:M) = varmmTT(m,1:M) + (TT3(m)-TT4(m))^2 + (TT3(1:M)-TT4(1:M)).^2;
        varmmDD(m,1:M) = varmmDD(m,1:M) + (DD3(m)-DD4(m))^2 + (DD3(1:M)-DD4(1:M)).^2;
        varmmSS(m,1:M) = varmmSS(m,1:M) + (SS3(m)-SS4(m))^2 + (SS3(1:M)-SS4(1:M)).^2;
    end
                    
    varfTT = varfTT + (TT1-TT4).*(TT3-TT2);
    varfDD = varfDD + (DD1-DD4).*(DD3-DD2);
    varfSS = varfSS + (SS1-SS4).*(SS3-SS2);

    varvTT = varvTT + (TT1-TT3).*(TT4-TT2);
    varvDD = varvDD + (DD1-DD3).*(DD4-DD2);
    varvSS = varvSS + (SS1-SS3).*(SS4-SS2);
 
    for m = 1:M
        var2fTT(m,1:M) = var2fTT(m,1:M) + (TT3(m)-TT3(1:M)).*(TT4(1:M)-TT4(m));
        var2fDD(m,1:M) = var2fDD(m,1:M) + (DD3(m)-DD3(1:M)).*(DD4(1:M)-DD4(m));
        var2fSS(m,1:M) = var2fSS(m,1:M) + (SS3(m)-SS3(1:M)).*(SS4(1:M)-SS4(m));
    end
    
end

Lactual = LL;
      
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Calculate variances
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

meanTT = sqrt(meanTT./(2*Lactual));
meanDD = sqrt(meanDD./(2*Lactual));
meanSS = sqrt(meanSS./(2*Lactual));

varmTT  = varmTT./(4*Lactual);
varmDD  = varmDD./(4*Lactual);
varmSS  = varmSS./(4*Lactual);

varmmTT = varmmTT./(4*Lactual);
varmmDD = varmmDD./(4*Lactual);
varmmSS = varmmSS./(4*Lactual);

varfTT  = varfTT.*(varfTT > 0)./(2*Lactual);
varfDD  = varfDD.*(varfDD > 0)./(2*Lactual);
varfSS  = varfSS.*(varfSS > 0)./(2*Lactual);

varvTT  = varvTT.*(varvTT > 0)./(2*Lactual);
varvDD  = varvDD.*(varvDD > 0)./(2*Lactual);
varvSS  = varvSS.*(varvSS > 0)./(2*Lactual);

var2fTT = var2fTT.*(var2fTT > 0)./(2*Lactual);
var2fDD = var2fDD.*(var2fDD > 0)./(2*Lactual);
var2fSS = var2fSS.*(var2fSS > 0)./(2*Lactual);