% MAIN FILE (ROSA-DV)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%   H.-X. Zhang, W.P. Dempsey, Jr., and J. Goutsias    %%%%%%%%%%%
%%%%%%%%%%        Probabilistic Sensitivity Analysis of         %%%%%%%%%%%
%%%%%%%%%%            Biochemical Reaction Systems              %%%%%%%%%%%
%%%%%%%%%%  The Journal of Chemical Physics, 131, 094101, 2009  %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%         Reaction-Oriented Sensitivity Analysis       %%%%%%%%%%%
%%%%%%%%%%             Derivative-Based Approximation           %%%%%%%%%%%
%%%%%%%%%%                Last Modified: 08/11/2009             %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

clear all;          % clear all parameters
close all;          % close all files

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Input parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

ncpu  = 2;          % number of computer processors

delta = 0.01;       % size of finite-difference
theta = 0.1;        % sensitivity threshold
time  = 360;        % simulation time (min)

iode = 0;           % iode=0 use of Matlab ODE solver (slow)
                    % iode=1 use of custom ODE solver (fast)

parameters;         % system parameters
q0_vec;             % initial molecular numbers
species = 19;       % number of species for sensitivity analysis

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Initialize
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

durtol  = 0.05;                                     % duration tolerance
tresol  = 1/60;                                     % time resolution (in min)
tpoints = (0:60*tresol:60*time);                    % time points for ODE integration
odetol  = 1e-6;                                     % tolerance for ODE solver
options = odeset('RelTol',odetol,'AbsTol',odetol);  % options for Matlab ODE solver

N         = size(q0,2);
tmax      = size(tpoints,2);
qf        = repmat(0,[tmax,N,M]);
qb        = repmat(0,[tmax,N,M]);
reac(1:M) = 1:M;
ndata     = time/tresol;
timings   = [0,60*time,ndata];
    
tcpu = cputime;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Solve system of ODEs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

for m = 1:M
    
    YR      = zeros(1,M);
    YR(m)   = delta;
    temp    = repmat(YR,[2 1]);      % replicate each element of YR twice
    YR      = temp(:).';             % replicate each element of YR twice
    kappaf  = kappa.*exp(+YR)';
    kappab  = kappa.*exp(-YR)';
    
    if iode == 1
        [t,qf(:,:,m)] = mapk(kappaf,q0,[odetol,odetol],timings);
        [t,qb(:,:,m)] = mapk(kappab,q0,[odetol,odetol],timings);
    else
        [t,qf(:,:,m)] = ode15s(@netode,tpoints,q0,options,N,kappaf,stoich);
        [t,qb(:,:,m)] = ode15s(@netode,tpoints,q0,options,N,kappab,stoich);
    end
    
end

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Approximate SESIs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

senseit;

tcpu = cputime - tcpu;
tcpu = tcpu/ncpu;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Plot SESI values
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

plotit;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Write sensitivity results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

writeit;