% MC-LHS variance estimation (ROSOSA-VSI)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%   H.-X. Zhang, W.P. Dempsey, Jr., and J. Goutsias    %%%%%%%%%%%
%%%%%%%%%%        Probabilistic Sensitivity Analysis of         %%%%%%%%%%%
%%%%%%%%%%            Biochemical Reaction Systems              %%%%%%%%%%%
%%%%%%%%%%  The Journal of Chemical Physics, 131, 094101, 2009  %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%  Reaction and Species-Oriented Sensitivity Analysis  %%%%%%%%%%%
%%%%%%%%%%                   VSI COMPUTATION                    %%%%%%%%%%%
%%%%%%%%%%              Last Modified: 08/11/2009               %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Initializations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

perm1 = repmat(0,[MN,L]);
perm2 = repmat(0,[MN,L]);
YRS1  = repmat(0,[MN,1]);
YRS2  = repmat(0,[MN,1]);
YRS3  = repmat(0,[MN,MN]);
YRS4  = repmat(0,[MN,MN]);

meanTT(1:MN) = 0;
meanDD(1:MN) = 0;
meanSS(1:MN) = 0;

varmTT(1:MN) = 0;
varmDD(1:MN) = 0;
varmSS(1:MN) = 0;

varmmTT(1:MN,1:MN) = 0;
varmmDD(1:MN,1:MN) = 0;
varmmSS(1:MN,1:MN) = 0;

varfTT(1:MN) = 0;
varfDD(1:MN) = 0;
varfSS(1:MN) = 0;

varvTT(1:MN) = 0;
varvDD(1:MN) = 0;
varvSS(1:MN) = 0;

var2fTT(1:MN,1:MN) = 0;
var2fDD(1:MN,1:MN) = 0;
var2fSS(1:MN,1:MN) = 0;

rand('state',sum(100*clock));  % initialize random number generator

constr(1:2*M,1) = (k*T)/h;     % standard chemical potential variation constraint

LL = L;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% MC-LHS iterations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Determine LHS permutations
%

for mn = 1:MN
    perm1(mn,:) = randperm(L);
    perm2(mn,:) = randperm(L);
end

for iter = 1:L
    
    act = 0;
    iterleft = L+1-iter;
    fprintf('Remaining MC-LHS iterations: %6.0f\n',iterleft)
    
%
% Calculate Latin hypercube samples
%

    uniform = rand(MN,1);
    index   = (perm1(:,iter)-1+uniform)./L;
    YR1     = rlambda*(-sqrt(2)*erfcinv(2*index(1:M)));       % inverse normal CDF
    YS1     = slambda*(-sqrt(2)*erfcinv(2*index((M+1):MN)));  % inverse normal CDF
    YRS1    = [YR1;YS1];
   
    temp    = repmat(YR1',[2 1]);     % replicate each element of YR1 twice
    YR1     = temp(:);                % replicate each element of YR1 twice
    svyn    = (-stoich.*(stoich < 0))'*YS1;
    kappa1  = kappa.*exp(-YR1).*exp(svyn);
    
    if kappa1 > constr
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end
    
    uniform = rand(MN,1);
    index   = (perm2(:,iter)-1+uniform)./L;
    YR2     = rlambda*(-sqrt(2)*erfcinv(2*index(1:M)));       % inverse normal CDF
    YS2     = slambda*(-sqrt(2)*erfcinv(2*index((M+1):MN)));  % inverse normal CDF
    YRS2    = [YR2;YS2];
   
    temp    = repmat(YR2',[2 1]);     % replicate each element of YR2 twice
    YR2     = temp(:);                % replicate each element of YR2 twice
    svyn    = (-stoich.*(stoich < 0))'*YS2;
    kappa2  = kappa.*exp(-YR2).*exp(svyn);
    
    if kappa2 > constr
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end
    
    for mn = 1:MN
        YRS3(mn,:)  = YRS2';
        YRS3(mn,mn) = YRS1(mn);
        YRS4(mn,:)  = YRS1';
        YRS4(mn,mn) = YRS2(mn);
    end
    
%
% Calculate response characteristics (timing, duration, and strength)
%    
% For kappa1

    if iode == 1
        [t,q] = mapk(kappa1,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa1,stoich);
    end
    [TT1,DD1,SS1] = featureit(species,q,t,tresol,durtol);
    
% For kappa2

    if iode == 1
        [t,q] = mapk(kappa2,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa2,stoich);
    end
    [TT2,DD2,SS2] = featureit(species,q,t,tresol,durtol);
    
% For kappa3 & kappa4

    for mn = 1:MN
        
        YR3    = YRS3(mn,1:M);
        YS3    = YRS3(mn,(M+1):MN)';
        temp   = repmat(YR3,[2 1]);      % replicate each element of YR3 twice
        YR3    = temp(:);                % replicate each element of YR3 twice
        svyn   = (-stoich.*(stoich < 0))'*YS3;
        kappa3 = kappa.*exp(-YR3).*exp(svyn);
      
        if kappa3 > constr
            fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
            LL = LL-1;
            act = 1;
            break
        end
        
        if iode == 1
            [t,q] = mapk(kappa3,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa3,stoich);
        end
        [TT3(mn),DD3(mn),SS3(mn)] = featureit(species,q,t,tresol,durtol);

        YR4    = YRS4(mn,1:M);
        YS4    = YRS4(mn,(M+1):MN)';
        temp   = repmat(YR4,[2 1]);      % replicate each element of YR4 twice
        YR4    = temp(:);                % replicate each element of YR4 twice
        svyn   = (-stoich.*(stoich < 0))'*YS4;
        kappa4 = kappa.*exp(-YR4).*exp(svyn);
      
        if kappa4 > constr
            fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
            LL = LL-1;
            act = 1;
            break
        end
        
        if iode == 1
           [t,q] = mapk(kappa4,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa4,stoich);
        end
        [TT4(mn),DD4(mn),SS4(mn)] = featureit(species,q,t,tresol,durtol);
        
    end
    
    if(act == 1)
        continue
    end
    
%
% Update mean and variance calculations
%
    
    meanTT = meanTT + TT1*TT2 + TT3.*TT4;
    meanDD = meanDD + DD1*DD2 + DD3.*DD4;
    meanSS = meanSS + SS1*SS2 + SS3.*SS4;
    
    varmTT = varmTT + (TT1-TT2)^2 + (TT3-TT4).^2;
    varmDD = varmDD + (DD1-DD2)^2 + (DD3-DD4).^2;
    varmSS = varmSS + (SS1-SS2)^2 + (SS3-SS4).^2;
    
    for mn = 1:MN
        varmmTT(mn,1:MN) = varmmTT(mn,1:MN)+(TT3(mn)-TT4(mn))^2+(TT3(1:MN)-TT4(1:MN)).^2;
        varmmDD(mn,1:MN) = varmmDD(mn,1:MN)+(DD3(mn)-DD4(mn))^2+(DD3(1:MN)-DD4(1:MN)).^2;
        varmmSS(mn,1:MN) = varmmSS(mn,1:MN)+(SS3(mn)-SS4(mn))^2+(SS3(1:MN)-SS4(1:MN)).^2;
    end
                    
    varfTT = varfTT + (TT1-TT4).*(TT3-TT2);
    varfDD = varfDD + (DD1-DD4).*(DD3-DD2);
    varfSS = varfSS + (SS1-SS4).*(SS3-SS2);

    varvTT = varvTT + (TT1-TT3).*(TT4-TT2);
    varvDD = varvDD + (DD1-DD3).*(DD4-DD2);
    varvSS = varvSS + (SS1-SS3).*(SS4-SS2);
 
    for mn = 1:MN
        var2fTT(mn,1:MN) = var2fTT(mn,1:MN)+(TT3(mn)-TT3(1:MN)).*(TT4(1:MN)-TT4(mn));
        var2fDD(mn,1:MN) = var2fDD(mn,1:MN)+(DD3(mn)-DD3(1:MN)).*(DD4(1:MN)-DD4(mn));
        var2fSS(mn,1:MN) = var2fSS(mn,1:MN)+(SS3(mn)-SS3(1:MN)).*(SS4(1:MN)-SS4(mn));
    end
   
end

Lactual = LL;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Calculate variances
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

meanTT  = sqrt(meanTT./(2*Lactual));
meanDD  = sqrt(meanDD./(2*Lactual));
meanSS  = sqrt(meanSS./(2*Lactual));

varmTT  = varmTT./(4*Lactual);
varmDD  = varmDD./(4*Lactual);
varmSS  = varmSS./(4*Lactual);

varmmTT = varmmTT./(4*Lactual);
varmmDD = varmmDD./(4*Lactual);
varmmSS = varmmSS./(4*Lactual);

varfTT  = varfTT.*(varfTT > 0)./(2*Lactual);
varfDD  = varfDD.*(varfDD > 0)./(2*Lactual);
varfSS  = varfSS.*(varfSS > 0)./(2*Lactual);

varvTT  = varvTT.*(varvTT > 0)./(2*Lactual);
varvDD  = varvDD.*(varvDD > 0)./(2*Lactual);
varvSS  = varvSS.*(varvSS > 0)./(2*Lactual);

var2fTT = var2fTT.*(var2fTT > 0)./(2*Lactual);
var2fDD = var2fDD.*(var2fDD > 0)./(2*Lactual);
var2fSS = var2fSS.*(var2fSS > 0)./(2*Lactual);