% MC-LHS variance estimation (SOSA-VSI)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%   H.-X. Zhang, W.P. Dempsey, Jr., and J. Goutsias    %%%%%%%%%%%
%%%%%%%%%%        Probabilistic Sensitivity Analysis of         %%%%%%%%%%%
%%%%%%%%%%            Biochemical Reaction Systems              %%%%%%%%%%%
%%%%%%%%%%  The Journal of Chemical Physics, 131, 094101, 2009  %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%         Species-Oriented Sensitivity Analysis        %%%%%%%%%%%
%%%%%%%%%%                   VSI COMPUTATION                    %%%%%%%%%%%
%%%%%%%%%%              Last Modified: 08/11/2009               %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Initializations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

perm1 = repmat(0,[N,L]);
perm2 = repmat(0,[N,L]);
YS1   = repmat(0,[N,1]);
YS2   = repmat(0,[N,1]);
YS3   = repmat(0,[N,N]);
YS4   = repmat(0,[N,N]);

meanTT(1:N) = 0;
meanDD(1:N) = 0;
meanSS(1:N) = 0;

varmTT(1:N) = 0;
varmDD(1:N) = 0;
varmSS(1:N) = 0;

varmmTT(1:N,1:N) = 0;
varmmDD(1:N,1:N) = 0;
varmmSS(1:N,1:N) = 0;

varfTT(1:N) = 0;
varfDD(1:N) = 0;
varfSS(1:N) = 0;

varvTT(1:N) = 0;
varvDD(1:N) = 0;
varvSS(1:N) = 0;

var2fTT(1:N,1:N) = 0;
var2fDD(1:N,1:N) = 0;
var2fSS(1:N,1:N) = 0;

rand('state',sum(100*clock));  % initialize random number generator

constr(1:2*M,1) = (k*T)/h;     % standard chemical potential variation constraint

LL = L;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% MC-LHS iterations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Determine LHS permutations
%

for n = 1:N
    perm1(n,:) = randperm(L);
    perm2(n,:) = randperm(L);
end

for iter = 1:L
    
    act = 0;
    iterleft = L+1-iter;
    fprintf('Remaining MC-LHS iterations: %6.0f\n',iterleft)
    
%
% Calculate Latin hypercube samples
%

    uniform = rand(N,1);
    index   = (perm1(:,iter)-1+uniform)./L;
    YS1     = slambda*(-sqrt(2)*erfcinv(2*index));     % inverse normal CDF
    svyn    = (-stoich.*(stoich < 0))'*YS1;
    kappa1  = kappa.*exp(svyn);
    
    if kappa1 > constr
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end
    
    uniform = rand(N,1);
    index   = (perm2(:,iter)-1+uniform)./L;
    YS2     = slambda*(-sqrt(2)*erfcinv(2*index));     % inverse normal CDF
    svyn    = (-stoich.*(stoich < 0))'*YS2;
    kappa2  = kappa.*exp(svyn);
 
    if kappa2 > constr
        fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
        LL = LL-1;
        continue
    end
    
    for n = 1:N
        YS3(n,:) = YS2';
        YS3(n,n) = YS1(n);
        YS4(n,:) = YS1';
        YS4(n,n) = YS2(n);
    end
    
%
% Calculate response characteristics (timing, duration, and strength)
%
% For kappa1
    
    if iode == 1
        [t,q] = mapk(kappa1,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa1,stoich);
    end
    [TT1,DD1,SS1] = featureit(species,q,t,tresol,durtol);
    
% For kappa2
    
    if iode == 1
        [t,q] = mapk(kappa2,q0,[odetol,odetol],timings);
    else
        [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa2,stoich);
    end
    [TT2,DD2,SS2] = featureit(species,q,t,tresol,durtol);
    
% For kappa3 & kappa4
    
    for n = 1:N
        
        svyn   = (-stoich.*(stoich < 0))'*YS3(n,:)';
        kappa3 = kappa.*exp(svyn);
        
        if kappa3 > constr
            fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
            LL  = LL-1;
            act = 1;
            break
        end
        
        if iode == 1
            [t,q] = mapk(kappa3,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa3,stoich);
        end
        [TT3(n),DD3(n),SS3(n)] = featureit(species,q,t,tresol,durtol);

        svyn   = (-stoich.*(stoich < 0))'*YS4(n,:)';
        kappa4 = kappa.*exp(svyn);
    
        if kappa4 > constr
            fprintf('**** WARNING: Standard chemical potential constraint is not satisfied\n')
            LL  = LL-1;
            act = 1;
            break
        end
        
        if iode == 1
            [t,q] = mapk(kappa4,q0,[odetol,odetol],timings);
        else
            [t,q] = ode15s(@netode,tpoints,q0,options,N,kappa4,stoich);
        end
        [TT4(n),DD4(n),SS4(n)] = featureit(species,q,t,tresol,durtol);
    
    end
    
    if(act == 1)
        continue
    end
    
    %
    % Update mean and variance calculations
    %
    
    meanTT = meanTT + TT1*TT2 + TT3.*TT4;
    meanDD = meanDD + DD1*DD2 + DD3.*DD4;
    meanSS = meanSS + SS1*SS2 + SS3.*SS4;
    
    varmTT = varmTT + (TT1-TT2)^2 + (TT3-TT4).^2;
    varmDD = varmDD + (DD1-DD2)^2 + (DD3-DD4).^2;
    varmSS = varmSS + (SS1-SS2)^2 + (SS3-SS4).^2;
    
    for n = 1:N
        varmmTT(n,1:N) = varmmTT(n,1:N) + (TT3(n)-TT4(n))^2 + (TT3(1:N)-TT4(1:N)).^2;
        varmmDD(n,1:N) = varmmDD(n,1:N) + (DD3(n)-DD4(n))^2 + (DD3(1:N)-DD4(1:N)).^2;
        varmmSS(n,1:N) = varmmSS(n,1:N) + (SS3(n)-SS4(n))^2 + (SS3(1:N)-SS4(1:N)).^2;
    end
                    
    varfTT = varfTT + (TT1-TT4).*(TT3-TT2);
    varfDD = varfDD + (DD1-DD4).*(DD3-DD2);
    varfSS = varfSS + (SS1-SS4).*(SS3-SS2);

    varvTT = varvTT + (TT1-TT3).*(TT4-TT2);
    varvDD = varvDD + (DD1-DD3).*(DD4-DD2);
    varvSS = varvSS + (SS1-SS3).*(SS4-SS2);
 
    for n = 1:N
        var2fTT(n,1:N) = var2fTT(n,1:N) + (TT3(n)-TT3(1:N)).*(TT4(1:N)-TT4(n));
        var2fDD(n,1:N) = var2fDD(n,1:N) + (DD3(n)-DD3(1:N)).*(DD4(1:N)-DD4(n));
        var2fSS(n,1:N) = var2fSS(n,1:N) + (SS3(n)-SS3(1:N)).*(SS4(1:N)-SS4(n));
    end
   
end

Lactual = LL;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Calculate variances
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

meanTT  = sqrt(meanTT./(2*Lactual));
meanDD  = sqrt(meanDD./(2*Lactual));
meanSS  = sqrt(meanSS./(2*Lactual));

varmTT  = varmTT./(4*Lactual);
varmDD  = varmDD./(4*Lactual);
varmSS  = varmSS./(4*Lactual);

varmmTT = varmmTT./(4*Lactual);
varmmDD = varmmDD./(4*Lactual);
varmmSS = varmmSS./(4*Lactual);

varfTT  = varfTT.*(varfTT > 0)./(2*Lactual);
varfDD  = varfDD.*(varfDD > 0)./(2*Lactual);
varfSS  = varfSS.*(varfSS > 0)./(2*Lactual);

varvTT  = varvTT.*(varvTT > 0)./(2*Lactual);
varvDD  = varvDD.*(varvDD > 0)./(2*Lactual);
varvSS  = varvSS.*(varvSS > 0)./(2*Lactual);

var2fTT = var2fTT.*(var2fTT > 0)./(2*Lactual);
var2fDD = var2fDD.*(var2fDD > 0)./(2*Lactual);
var2fSS = var2fSS.*(var2fSS > 0)./(2*Lactual);